#!/usr/bin/python
import sys
import os

from csc_interface import *
from csc_convert import *
from csc_edit import *
from csc_globals import *

def main():
	tasks = parseInput()
	if debugMode: print tasks
	if os.path.exists("error.log"):
		os.remove("error.log")
	showedVersion = False
	while tasks != []:
		curTask = tasks[0]
		tasks = tasks[1:]
		if "interactive" in curTask[0]:
			if not showedVersion:
				showVersion()
				showedVersion = True
			curTask += (tasks,)
		try:
			functions[curTask[0]](*curTask[1:])
		except:
			if debugMode: raise
			else: tasks.append( ("error",sys.exc_info(),tasks) )
			#displayError(sys.exc_info(),tasks)
		if debugMode: print tasks

def displayError(x,tasks):
	global errnum
	from traceback import print_tb
	from exceptions import KeyboardInterrupt
	if x[0] == KeyboardInterrupt:
		sys.exit(1)
	print "An unknown error has occurred."
	print "This may be the result of bad user input, or a program bug."
	if "interactive" in tasks[0]:
		del tasks[1:]
	try:
		logfile = open("error.log",'a')
	except:
		print "Additionally, an error occurred when trying to open"
		print "the log file. The program will now terminate."
		sys.exit(1)
	if errnum == 0:
		print "If you want to report the error, it is error"
		print "number", errnum, "as recorded in error.log."
		print "Note that error.log is removed each time the"
		print "program is run."
	else:
		print "It is error", errnum, "in error.log."
	logfile.write("Error #" + `errnum` + '\n')
	logfile.write(' '.join(sys.argv) + '\n')
	try:
		logfile.write("Task: " + tasks[0][0] + '\n')
	except:
		logfile.write("Task: <error>")
	try:
		logfile.write("Error: " + str(x[0]) + '\n')
	except:
		logfile.write("Error: <error>")
	try:
		logfile.write("Message: " + str(x[1]) + '\n')
	except:
		logfile.write("Message: <error>")
	print_tb(x[2],1,logfile)
	logfile.write("\n\n")
	logfile.close()
	errnum += 1

def parseInput():
	"""Get command-line input"""
	tasks = []
	currentArg = ""
	if len(sys.argv) == 1:
		tasks.append( ("interactive",) )
	i = 1
	curFile = ""
	while i < len(sys.argv):
		if sys.argv[i] == "--version":
			showVersion()
			sys.exit()
		if sys.argv[i] == '-?' or sys.argv[i] == '-h' or sys.argv[i] == '/?':
			displayUsage(sys.argv[0],sys.platform)
			sys.exit()
		if sys.argv[i] == '-?x' or sys.argv[i] == '-hx':
			advancedHelp()
			sys.exit()
		if sys.argv[i] == '-?w' or sys.argv[i] == '-hw':
			weaponHelp()
			sys.exit()
		if sys.argv[i][1] == '%':
			bkup = sys.argv[i][0]
			curFile = sys.argv[i][2:]
			if bkup == 'b': tasks.append( ("backup", curFile) )
			elif bkup != 'd':
				print "Invalid file flag."
				return
#			try:
#				open(curFile).close()
#			except:
#				print sys.exc_value
#				sys.exit()
			i += 1
#		 elif sys.argv[i] == '-e':
#			 x = ""
#			 try:
#				 if sys.argv[i+1] == '@': x = defPLPath
#				 else: x = sys.argv[i+1]
#				 tasks.append( ("extract-mac-profile",x,sys.argv[i+2]) )
#			 except:
#				 print "Error: Source or destination not specified for extraction."
#				 sys.exit()
#			 i += 3
#		 elif sys.argv[i] == '-s':
#			 try:
#				 if sys.argv[i+2] == '@': x = defPLPath
#				 else: x = sys.argv[i+2]
#				 tasks.append( ("store-mac-profile",sys.argv[i+1],x) )
#			 except:
#				 print "Error: Source or destination not specified for storing profile."
#				 sys.exit()
#			 i += 3
		elif sys.argv[i] == '-ec' or sys.argv[i] == '-e':
			try:
				if sys.argv[i+1] == '@': pl = defPLPath
				else: pl = sys.argv[i+1]
				try:
					open(pl).close()
				except:
					print sys.exc_value
					if pl == defPLPath:
						print "(Open Cave Story, then quit and try again.)"
					sys.exit()
				tasks.append( ("extract-mac-profile",pl,".pycs_temp") )
				tasks.append( ("convert-profile",".pycs_temp",curFile) )
				tasks.append( ("clean-temp-file",".pycs_temp") )
			except:
				print "Error: Source not specified for extraction."
				sys.exit()
			i += 2
		elif sys.argv[i] == '-cs' or sys.argv[i] == '-s':
			if curFile == "": specifyFileError()
			try:
				if sys.argv[i+1] == '@': pl = defPLPath
				else: pl = sys.argv[i+2]
				try:
					open(pl).close()
				except:
					print sys.exc_value
					if pl == defPLPath:
						print "(Open Cave Story, then quit and try again.)"
					sys.exit()
				tasks.append( ("convert-profile",curFile,".pycs_temp") )
				tasks.append( ("store-mac-profile",".pycs_temp",pl) )
				tasks.append( ("clean-temp-file",".pycs_temp") )
			except:
				print "Error: Destination not specified for storing profile."
				sys.exit()
			i += 2
		elif sys.argv[i][:2] == '-w':
			if curFile == "": specifyFileError()
			arg = sys.argv[i][2:]
			tasks.append( ("remove-weapon", arg, curFile) )
			i += 1
		elif sys.argv[i][:2] == '+w':
			# Syntax: +w4[@3[(12)]][=100[/100]] : weapon[@level[(energy)]][=ammo[/max]]
			# Note: Use [] instead of () since () are special characters in the shell
			if curFile == "": specifyFileError()
			args = sys.argv[i][2:]
			# Validate args
			validchars = ('a','b','c','d','e','f','g','h','i','j','k','l','m', 'n','o','p','q','r','s','t','u','v','w','x','y','z', '0','1','2','3','4','5','6','7','8','9','@','=','[',']','/')
			for x in args.lower():
				if x not in validchars:
					print "Syntax error in add weapon directive:", args
					if debugMode:
						print sys.exc_type
						print sys.exc_value
						print "Line", sys.exc_traceback.tb_lineno
					sys.exit()
			# Parse args
			if '@' in args:
				# Specified level
				args = args.split('@')
				if '=' in args[1]:
					# Specified ammo
					args[1:] = args[1].split('=')
				else:
					# Default ammo
					args.append("default")
			elif '=' in args:
				# Default level; specified ammo
				args = args.split('=')
				args = [args[0], "1", args[1]]
			else:
				# Default level; default ammo
				args = [args, "1", "default"]
			#try:
			#	args = args.split('@')
			#	try:
			#		args[1:] = args[1].split('=')
			#	except:
			#		args.append("default")
			#except:
			#	args.append("1")
			#	args.append("default")
			if debugMode: print args
			task = ("add-weapon",)
			for x in args:
				if debugMode: print x
				task += (x,)
			task += (curFile,)
			if debugMode: print task
			tasks.append( task )
			#except:
			#	print "Syntax error in add weapon directive."
			#	if debugMode:
			#		print sys.exc_type
			#		print sys.exc_value
			#		print "Line", sys.exc_traceback.tb_lineno
			#	sys.exit()
			i += 1
		elif sys.argv[i][:2] == '*w':
			if curFile == "": specifyFileError()
			arg = sys.argv[i][2:]
			tasks.append( ("equip-weapon", arg, curFile) )
			i += 1
		elif sys.argv[i][:2] == '-t':
			if curFile == "": specifyFileError()
			arg = sys.argv[i][2:]
			tasks.append( ("remove-teleporter", arg, curFile) )
			i += 1
		elif sys.argv[i][:2] == '+t':
			# Syntax: +t4@4 : menu@location
			if curFile == "": specifyFileError()
			args = sys.argv[i][2:]
			try:
				args = args.split('@')
				if debugMode: print args
				task = ("add-teleporter",)
				for x in args:
					if debugMode: print x
					task += (x,)
				task += (curFile,)
				if debugMode: print task
				tasks.append( task )
			except:
				print "Syntax error in add teleport location directive."
				if debugMode:
					print sys.exc_type
					print sys.exc_value
					print "Line", sys.exc_traceback.tb_lineno
				sys.exit()
			i += 1
		elif sys.argv[i][:2] == '-i':
			if curFile == "": specifyFileError()
			arg = sys.argv[i][2:]
			tasks.append( ("remove-item", int(arg), curFile) )
			i += 1
		elif sys.argv[i][:2] == '+i':
			if curFile == "": specifyFileError()
			arg = sys.argv[i][2:]
			tasks.append( ("add-item", int(arg), curFile) )
			i += 1
		elif sys.argv[i][:2] == '*i':
			#syntax *i4=on
			if curFile == "": specifyFileError()
			try:
				args = sys.argv[i][2:]
				args = args.split('=')
				if args[1].lower() == 'on':
					args[1] = True
				elif args[1].lower() == 'off':
					args[1] = False
				else:
					raise TypeError # Why TypeError?
			except:
				print "Syntax error in equip item directive."
				sys.exit()
			tasks.append( ("equip-item", int(args[0]), args[1], curFile) )
			i += 1
		elif sys.argv[i][:2] == '-f':
			if curFile == "": specifyFileError()
			arg = sys.argv[i][2:]
			tasks.append( ("unset-flag", int(arg), curFile) )
			i += 1
		elif sys.argv[i][:2] == '+f':
			if curFile == "": specifyFileError()
			arg = sys.argv[i][2:]
			tasks.append( ("set-flag", int(arg), curFile) )
			i += 1
		elif sys.argv[i][:2] == '-v':
			if curFile == "": specifyFileError()
			if len(sys.argv[i]) == 2: parts = "bwitf"
			else: parts = sys.argv[i][2:]
			tasks.append( ("view", parts, curFile) )
			i += 1
		elif sys.argv[i][:2] == '-m':
			# Syntax: -m horz vert
			# (This is for relative movement.)
			if curFile == "": specifyFileError()
			try:
				args = sys.argv[2:].split(',')
				tasks.append( ("move", args[0], args[1], curFile) )
			except:
				print "Invalid move directive."
				return
			i += 1
		elif sys.argv[i] == '-x':
			if curFile == "": specifyFileError()
			i += 1
			attrs = {}
			try:
				while sys.argv[i] != '-':
					arg = sys.argv[i].split('=')
					try:
						attrs[arg[0]] = int(arg[1])
					except:
						attrs[arg[0]] = arg[1]
					i += 1
			except:
				print "Advanced editing section was not terminated by -"
				sys.exit()
			if attrs != {}:
				tasks.append( ("advanced-edit", attrs, curFile) )
			i += 1
		else:
			print "Invalid argument %s" % sys.argv[i]
			sys.exit()
	return tasks
#	 if len(sys.argv) > 1:
#		 if sys.argv[1] == "-?" or sys.argv[1] == "-h" or sys.argv[1] == "-h":
#			 print "Usage:", sys.argv[0], "name [opt] [source [dest]]"
#			 sys.exit()
#		 elif sys.argv[1] == "-e":
#			 mode = "extract-mac-profile"
#			 source = sys.argv[2] if len(sys.argv) > 2 else "com.nakiwo.doukutsu.plist"
#			 if sys.platform == "darwin" and '/' not in source and '\\' not in source:
#				 source = "~/Library/Preferences/" + source;
#			 dest = sys.argv[3] if len(sys.argv) > 3 else "profile.dsv"
#		 elif sys.argv[1] == "-s":
#			 mode = "store-mac-profile"
#			 source = sys.argv[2] if len(sys.argv) > 2 else "profile.dsv"
#			 dest = sys.argv[3] if len(sys.argv) > 3 else defPLPath
#		 elif sys.argv[1] == "-m":
#			 mode = "win-to-mac"
#			 source = sys.argv[2] if len(sys.argv) > 2 else "profile.dat"
#			 dest = sys.argv[3] if len(sys.argv) > 3 else "profile.dsv"
#		 elif sys.argv[1] == "-w":
#			 mode = "mac-to-win"
#			 source = sys.argv[2] if len(sys.argv) > 2 else "profile.dsv"
#			 dest = sys.argv[3] if len(sys.argv) > 3 else "profile.dat"
#		 else:
#			 source = sys.argv[2]
#			 if len(sys.argv) > 3: dest = sys.argv[3]
#	 return mode

def view(parts, src):
	profile = open(src,"rb")
	print "Saved Game", profile.name
	print ""
	profile.seek(0x08)
	map = numFromLittleEndian(profile.read(4))
	song = numFromLittleEndian(profile.read(4))
	profile.seek(1,os.SEEK_CUR)
	horz = numFromLittleEndian(profile.read(2))
	profile.seek(2,os.SEEK_CUR)
	vert = numFromLittleEndian(profile.read(2))
	profile.seek(1,os.SEEK_CUR)
	dir = numFromLittleEndian(profile.read(1))
	profile.seek(3,os.SEEK_CUR)
	maxhp = numFromLittleEndian(profile.read(2))
	star = numFromLittleEndian(profile.read(1))
	profile.seek(1,os.SEEK_CUR)
	hp = numFromLittleEndian(profile.read(2))
	profile.seek(2,os.SEEK_CUR)
	curweap = numFromLittleEndian(profile.read(4))
	profile.seek(4,os.SEEK_CUR)
	equip = numFromLittleEndian(profile.read(4))
	profile.seek(4,os.SEEK_CUR)
	time = numFromLittleEndian(profile.read(4))
	if 'b' in parts:
		horz /= 32.0
		vert /= 32.0
		print "Quote is in map", map, "at", '(' + `horz` + ',' + `vert` + ')', "facing", "right" if dir == 2 else "left"
		print "Current song:", song
		print "Health:", `hp` + '/' + `maxhp`
		if equip & 256: print "Timer:", time
		print ""
	if 'w' in parts: print "WEAPONS"
	for i in range(0,8):
		type = numFromLittleEndian(profile.read(4))
		level = numFromLittleEndian(profile.read(4))
		energy = numFromLittleEndian(profile.read(4))
		maxammo = numFromLittleEndian(profile.read(4))
		curammo = numFromLittleEndian(profile.read(4))
		if 'w' in parts and type != 0:
			print "Weapon", type, "at level", level, "(energy", `energy` + ')', "ammo", `curammo` + '/' + `maxammo`, "(equipped)" if i == equip else ""
	if 'w' in parts:
		if equip & 128:
			print "Whimsical Stars:", star
		print ""
	if 'i' in parts: print "ITEMS"
	
	for i in range(0,32):
		type = numFromLittleEndian(profile.read(4))
		if 'i' in parts and type != 0:
			print "Item", type
	if 'i' in parts:
		for i in (1,2,4,8,16,32,64,128,256):
			if equip & i:
				print "Equipped", i
		print ""
	if 't' in parts: print "TELEPORTER"
	for i in range(0,8):
		menu = numFromLittleEndian(profile.read(4))
		evt = numFromLittleEndian(profile.read(4))
		if 't' in parts and menu != 0:
			print "Teleporter menu option", menu, "goes to event", `evt` + '.'
	if 't' in parts: print ""
	if 'f' in parts:
		print "FLAGS"
		flagblock = 0
		onthisline = 0
		for i in range(0,8000):
			offset = (i / 8) + 0x21C
			mask = 2 ** (i % 8)
			if mask == 1:
				flagblock = numFromLittleEndian(profile.read(4))
			if flagblock & mask:
				if onthisline > 8:
					onthisline = 0
					sys.stdout.write('\n')
				sys.stdout.write("%04d" % i)
				sys.stdout.write('    ')
				onthisline += 1
		print ""
	profile.close()

def bkup(file):
	bkupfile = file.split('.')
	bkupfile[0] += "-backup"
	exists = lambda f: os.path.exists('.'.join(f))
	if exists(bkupfile):
		bkupfile.insert(1,'1')
		while exists(bkupfile):
			n = int(bkupfile[1])
			n += 1
			bkupfile[1] = str(n)
	bkupfile = '.'.join(bkupfile)
	print "Saving backup of %s to %s." % (file,bkupfile)
	orig = open(file,'rb')
	new = open(bkupfile,'wb')
	new.write(orig.read())
	new.close()
	orig.close()
	tasks.append( ("advanced-interactive",) )

functions = {"backup":bkup, "extract-mac-profile":extract, "store-mac-profile":store,
			 "convert-profile":convert, "view":view, "remove-weapon":wrem,
			 "add-weapon":wadd, "equip-weapon":wequip, "remove-item":irem, "add-item":iadd,
			 "equip-item":iequip, "set-flag":fset, "unset-flag":fclear,
			 "remove-teleporter":trem, "add-teleporter":tadd, "advanced-edit":edit,
			 "move":move, "interactive":getUserInput, "clean-temp-file":deleteTemp,
			 "advanced-interactive":interactiveEditMode, "choose-file":chooseMainFile,
			 "interactive-view":interactiveView, "error":displayError}

main()
